@@ -26,10 +26,15 @@ ARMOR_FIRE_NODES = {
 	{"ethereal:crystal_spike",  2, 1},
 	{"ethereal:fire_flower",    2, 1},
 	{"default:torch",           1, 1},
+	{"default:torch_ceiling",   1, 1},
+	{"default:torch_wall",      1, 1},
 }
 
 local skin_mod = nil
 local inv_mod = nil
+local use_player_monoids = minetest.global_exists("player_monoids")
+local use_armor_monoid = minetest.global_exists("armor_monoid")
+local preview_textures = {}
 
 local modpath = minetest.get_modpath(ARMOR_MOD_NAME)
 local worldpath = minetest.get_worldpath()
@@ -56,31 +61,42 @@ armor = {
 	timer = 0,
 	elements = {"head", "torso", "legs", "feet"},
 	physics = {"jump","speed","gravity"},
-	formspec = "size[8,8.5]image[2,0.75;2,4;armor_preview]"
-		.."list[current_player;main;0,4.5;8,4;]"
-		.."list[current_player;craft;4,1;3,3;]"
-		.."list[current_player;craftpreview;7,2;1,1;]"
-		.."listring[current_player;main]"
-		.."listring[current_player;craft]",
+	formspec = "size[8,8.5]"..
+		default.gui_bg..
+		default.gui_bg_img..
+		default.gui_slots..
+		default.get_hotbar_bg(0,4.25)..
+		"image[2,0.5;2,4;armor_preview]"..
+		"list[current_player;main;0,4.25;8,1;]"..
+		"list[current_player;main;0,5.5;8,3;8]"..
+		"list[current_player;craft;4,0.5;3,3;]"..
+		"list[current_player;craftpreview;7,1.5;1,1;]"..
+		"listring[current_player;main]"..
+		"listring[current_player;craft]",
 	textures = {},
 	default_skin = "character",
-	version = "0.4.7",
+	version = "0.4.8",
 }
 
+local armor_formpage = "image[2.5,0;2,4;armor_preview]"..
+	default.gui_bg..
+	default.gui_bg_img..
+	default.gui_slots..
+	default.get_hotbar_bg(0,4.25)..
+	"label[5,1;Level: armor_level]"..
+	"label[5,1.5;Heal:  armor_heal]"..
+	"label[5,2;Fire:  armor_fire]"..
+	"label[5,2.5;Radiation:  armor_radiation]"..
+	"list[current_player;main;0,4.25;8,1;]"..
+	"list[current_player;main;0,5.5;8,3;8]"
 if minetest.get_modpath("inventory_plus") then
 	inv_mod = "inventory_plus"
-	armor.formspec = "size[8,8.5]button[0,0;2,0.5;main;Back]"
-		.."image[2.5,0.75;2,4;armor_preview]"
-		.."label[5,1;Level: armor_level]"
-		.."label[5,1.5;Heal:  armor_heal]"
-		.."label[5,2;Fire:  armor_fire]"
-		.."label[5,2.5;Radiation:  armor_radiation]"
-		.."list[current_player;main;0,4.5;8,4;]"
+	armor.formspec = "size[8,8.5]button[6,0;2,0.5;main;Back]"..armor_formpage
 	if minetest.get_modpath("crafting") then
 		inventory_plus.get_formspec = function(player, page)
 		end
 	end
-elseif minetest.get_modpath("unified_inventory") then
+elseif minetest.get_modpath("unified_inventory") and not unified_inventory.sfinv_compat_layer then
 	inv_mod = "unified_inventory"
 	unified_inventory.register_button("armor", {
 		type = "image",
@@ -90,31 +106,48 @@ elseif minetest.get_modpath("unified_inventory") then
 		get_formspec = function(player, perplayer_formspec)
 			local fy = perplayer_formspec.formspec_y
 			local name = player:get_player_name()
-			local formspec = "background[0.06,"..fy..";7.92,7.52;3d_armor_ui_form.png]"
-				.."label[0,0;Armor]"
-				.."list[detached:"..name.."_armor;armor;0,"..fy..";2,3;]"
-				.."image[2.5,"..(fy - 0.25)..";2,4;"..armor.textures[name].preview.."]"
-				.."label[5.0,"..(fy + 0.0)..";Level: "..armor.def[name].level.."]"
-				.."label[5.0,"..(fy + 0.5)..";Heal:  "..armor.def[name].heal.."]"
-				.."label[5.0,"..(fy + 1.0)..";Fire:  "..armor.def[name].fire.."]"
-				.."label[5.0,"..(fy + 1.5)..";Radiation:  "..armor.def[name].radiation.."]"
-				.."listring[current_player;main]"
-				.."listring[detached:"..name.."_armor;armor]"
+			local formspec = "background[0.06,"..fy..";7.92,7.52;3d_armor_ui_form.png]"..
+				"label[0,0;Armor]"..
+				"list[detached:"..name.."_armor;armor;0,"..fy..";2,3;]"..
+				"image[2.5,"..(fy - 0.25)..";2,4;"..armor.textures[name].preview.."]"..
+				"label[5.0,"..(fy + 0.0)..";Level: "..armor.def[name].level.."]"..
+				"label[5.0,"..(fy + 0.5)..";Heal:  "..armor.def[name].heal.."]"..
+				"label[5.0,"..(fy + 1.0)..";Fire:  "..armor.def[name].fire.."]"..
+				"label[5.0,"..(fy + 1.5)..";Radiation:  "..armor.def[name].radiation.."]"..
+				"listring[current_player;main]"..
+				"listring[detached:"..name.."_armor;armor]"
 			return {formspec=formspec}
 		end,
 	})
 elseif minetest.get_modpath("inventory_enhanced") then
 	inv_mod = "inventory_enhanced"
+elseif minetest.get_modpath("smart_inventory") then
+	inv_mod = "smart_inventory"
+elseif minetest.get_modpath("sfinv") then
+	inv_mod = "sfinv"
+	armor.formspec = armor_formpage
+	sfinv.register_page("3d_armor:armor", {
+		title = "Armor",
+		get = function(self, player, context)
+			local name = player:get_player_name()
+			local formspec = armor:get_armor_formspec(name, true)
+			return sfinv.make_formspec(player, context, formspec, false)
+		end
+	})
 end
 
-if minetest.get_modpath("skins") then
-	skin_mod = "skins"
-elseif minetest.get_modpath("simple_skins") then
-	skin_mod = "simple_skins"
-elseif minetest.get_modpath("u_skins") then
-	skin_mod = "u_skins"
-elseif minetest.get_modpath("wardrobe") then
-	skin_mod = "wardrobe"
+local skin_mods = {"skins", "u_skins", "simple_skins", "wardrobe"}
+for _, mod in pairs(skin_mods) do
+	local path = minetest.get_modpath(mod)
+	if path then
+		local dir_list = minetest.get_dir_list(path.."/textures")
+		for _, fn in pairs(dir_list) do
+			if fn:find("_preview.png$") then
+				preview_textures[fn] = true
+			end
+		end
+		skin_mod = mod
+	end
 end
 
 armor.def = {
@@ -153,7 +186,7 @@ armor.set_player_armor = function(self, player)
 	local textures = {}
 	local physics_o = {speed=1,gravity=1,jump=1}
 	local material = {type=nil, count=1}
-	local preview = armor:get_preview(name) or "character_preview.png"
+	local preview = armor:get_preview(name)
 	for _,v in ipairs(self.elements) do
 		elements[v] = false
 	end
@@ -214,8 +247,23 @@ armor.set_player_armor = function(self, player)
 		armor_groups.fleshy = 100 - armor_level
 		armor_groups.radiation = 100 - armor_radiation
 	end
-	player:set_armor_groups(armor_groups)
-	player:set_physics_override(physics_o)
+	if use_armor_monoid then
+		armor_monoid.monoid:add_change(player, {
+			fleshy = armor_groups.fleshy / 100
+		}, "3d_armor:armor")
+	else
+		player:set_armor_groups(armor_groups)
+	end
+	if use_player_monoids then
+		player_monoids.speed:add_change(player, physics_o.speed,
+			"3d_armor:physics")
+		player_monoids.jump:add_change(player, physics_o.jump,
+			"3d_armor:physics")
+		player_monoids.gravity:add_change(player, physics_o.gravity,
+			"3d_armor:physics")
+	else
+		player:set_physics_override(physics_o)
+	end
 	self.textures[name].armor = armor_texture
 	self.textures[name].preview = preview
 	self.def[name].state = state
@@ -233,7 +281,7 @@ end
 
 armor.update_armor = function(self, player)
 	-- Legacy support: Called when armor levels are changed
-	-- Other mods can hook on to this function, see hud mod for example 
+	-- Other mods can hook on to this function, see hud mod for example
 end
 
 armor.get_player_skin = function(self, name)
@@ -249,12 +297,14 @@ armor.get_player_skin = function(self, name)
 end
 
 armor.get_preview = function(self, name)
-	if skin_mod == "skins" then
-		return armor:get_player_skin(name).."_preview.png"
+	local preview = armor:get_player_skin(name).."_preview.png"
+	if preview_textures[preview] then
+		return preview
 	end
+	return "character_preview.png"
 end
 
-armor.get_armor_formspec = function(self, name)
+armor.get_armor_formspec = function(self, name, listring)
 	if not armor.textures[name] then
 		minetest.log("error", "3d_armor: Player texture["..name.."] is nil [get_armor_formspec]")
 		return ""
@@ -263,12 +313,17 @@ armor.get_armor_formspec = function(self, name)
 		minetest.log("error", "3d_armor: Armor def["..name.."] is nil [get_armor_formspec]")
 		return ""
 	end
-	local formspec = armor.formspec.."list[detached:"..name.."_armor;armor;0,1;2,3;]"
+	local formspec = armor.formspec.."list[detached:"..name.."_armor;armor;0,0.5;2,3;]"
+	if listring == true then
+		formspec = formspec.."listring[current_player;main]"..
+			"listring[detached:"..name.."_armor;armor]"
+	end
 	formspec = formspec:gsub("armor_preview", armor.textures[name].preview)
 	formspec = formspec:gsub("armor_level", armor.def[name].level)
 	formspec = formspec:gsub("armor_heal", armor.def[name].heal)
 	formspec = formspec:gsub("armor_fire", armor.def[name].fire)
 	formspec = formspec:gsub("armor_radiation", armor.def[name].radiation)
+	formspec = formspec:gsub("player_name", armor.def[name].radiation)
 	return formspec
 end
 
@@ -277,20 +332,33 @@ armor.update_inventory = function(self, player)
 	if not name or inv_mod == "inventory_enhanced" then
 		return
 	end
-	if inv_mod == "unified_inventory" then
+	if inv_mod == "smart_inventory" then
+		local state = smart_inventory.get_page_state("player", name)
+		if state then
+			state:get("update_hook"):submit()
+		end
+	elseif inv_mod == "sfinv" then
+		if sfinv.set_page then
+			sfinv.set_page(player, "3d_armor:armor")
+		else
+			-- Backwards compat
+			sfinv.set_player_inventory_formspec(player, {
+				page = "3d_armor:armor"
+			})
+		end
+	elseif inv_mod == "unified_inventory" then
 		if unified_inventory.current_page[name] == "armor" then
 			unified_inventory.set_inventory_formspec(player, "armor")
 		end
 	else
-		local formspec = armor:get_armor_formspec(name)
 		if inv_mod == "inventory_plus" then
-			formspec = formspec.."listring[current_player;main]"
-				.."listring[detached:"..name.."_armor;armor]"
+			local formspec = armor:get_armor_formspec(name, true)
 			local page = player:get_inventory_formspec()
 			if page:find("detached:"..name.."_armor") then
 				inventory_plus.set_inventory_formspec(player, formspec)
 			end
 		elseif not core.setting_getbool("creative_mode") then
+			local formspec = armor:get_armor_formspec(name)
 			player:set_inventory_formspec(formspec)
 		end
 	end
@@ -323,6 +391,13 @@ armor.get_valid_player = function(self, player, msg)
 	return name, player_inv, armor_inv, pos
 end
 
+armor.drop_armor = function(pos, stack)
+	local obj = minetest.add_item(pos, stack)
+	if obj then
+		obj:setvelocity({x=math.random(-1, 1), y=5, z=math.random(-1, 1)})
+	end
+end
+
 -- Register Player Model
 
 default.player_register_model("3d_armor_character.b3d", {
@@ -350,7 +425,7 @@ minetest.register_on_player_receive_fields(function(player, formname, fields)
 		return
 	end
 	if inv_mod == "inventory_plus" and fields.armor then
-		local formspec = armor:get_armor_formspec(name)
+		local formspec = armor:get_armor_formspec(name, true)
 		inventory_plus.set_inventory_formspec(player, formspec)
 		return
 	end
@@ -406,7 +481,7 @@ minetest.register_on_joinplayer(function(player)
 	for i=1, 6 do
 		local stack = player_inv:get_stack("armor", i)
 		armor_inv:set_stack("armor", i, stack)
-	end	
+	end
 	armor.def[name] = {
 		state = 0,
 		count = 0,
@@ -446,12 +521,14 @@ minetest.register_on_joinplayer(function(player)
 			armor.textures[name].skin = skin
 		end
 	end
-	if minetest.get_modpath("player_textures") then
-		local filename = minetest.get_modpath("player_textures").."/textures/player_"..name
-		local f = io.open(filename..".png")
-		if f then
-			f:close()
-			armor.textures[name].skin = "player_"..name..".png"
+	local texture_path = minetest.get_modpath("player_textures")
+	if texture_path then
+		local dir_list = minetest.get_dir_list(texture_path.."/textures")
+		for _, fn in pairs(dir_list) do
+			if fn == "player_"..name..".png" then
+				armor.textures[name].skin = fn
+				break
+			end
 		end
 	end
 	for i=1, ARMOR_INIT_TIMES do
@@ -465,12 +542,6 @@ minetest.register_on_joinplayer(function(player)
 end)
 
 if ARMOR_DROP == true or ARMOR_DESTROY == true then
-	armor.drop_armor = function(pos, stack)
-		local obj = minetest.add_item(pos, stack)
-		if obj then
-			obj:setvelocity({x=math.random(-1, 1), y=5, z=math.random(-1, 1)})
-		end
-	end
 	minetest.register_on_dieplayer(function(player)
 		local name, player_inv, armor_inv, pos = armor:get_valid_player(player, "[on_dieplayer]")
 		if not name then
@@ -496,21 +567,21 @@ if ARMOR_DROP == true or ARMOR_DESTROY == true then
 		end
 		if ARMOR_DESTROY == false then
 			minetest.after(ARMOR_BONES_DELAY, function()
-				local node = minetest.get_node(vector.round(pos))
-				if node then
-					if node.name ~= "bones:bones" then
-						pos.y = pos.y+1
-						node = minetest.get_node(vector.round(pos))
-						if node.name ~= "bones:bones" then
-							minetest.log("warning", "Failed to add armor to bones node.")
-							return
-						end
+				local meta = nil
+				local maxp = vector.add(pos, 8)
+				local minp = vector.subtract(pos, 8)
+				local bones = minetest.find_nodes_in_area(minp, maxp, {"bones:bones"})
+				for _, p in pairs(bones) do
+					local m = minetest.get_meta(p)
+					if m:get_string("owner") == name then
+						meta = m
+						break
 					end
-					local meta = minetest.get_meta(vector.round(pos))
-					local owner = meta:get_string("owner")
+				end
+				if meta then
 					local inv = meta:get_inventory()
 					for _,stack in ipairs(drop) do
-						if name == owner and inv:room_for_item("main", stack) then
+						if inv:room_for_item("main", stack) then
 							inv:add_item("main", stack)
 						else
 							armor.drop_armor(pos, stack)
